package dev.fitko.fitconnect.client.util;

import com.nimbusds.jose.JWEObject;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.metadata.*;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Purpose;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.*;
import dev.fitko.fitconnect.client.sender.model.AttachmentPayload;
import dev.fitko.fitconnect.client.sender.model.EncryptedSubmissionPayload;
import dev.fitko.fitconnect.client.sender.model.SubmissionPayload;
import dev.fitko.fitconnect.client.subscriber.model.DecryptedAttachmentPayload;
import dev.fitko.fitconnect.client.subscriber.model.ReceivedAttachment;

import java.io.File;
import java.net.URI;
import java.text.ParseException;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

public final class SubmissionUtil {

    private SubmissionUtil() {
    }

    private static CreateSubmission createSubmission(final UUID destinationId, final ServiceType serviceType, final List<UUID> attachmentIdsToAnnounce) {
        return CreateSubmission.builder()
                .destinationId(destinationId)
                .announcedAttachments(attachmentIdsToAnnounce)
                .serviceType(serviceType)
                .build();
    }

    public static PublicServiceType buildPublicServiceType(final ServiceType serviceType) {
        final var publicServiceType = new PublicServiceType();
        publicServiceType.setIdentifier(serviceType.getIdentifier());
        if (serviceType.getName() != null) {
            publicServiceType.setName(serviceType.getName());
        }
        if (serviceType.getDescription() != null) {
            publicServiceType.setDescription(serviceType.getDescription());
        }
        return publicServiceType;
    }

    public static Metadata buildMetadataToSend(final Data data, final ServiceType serviceType, final List<AttachmentPayload> encryptedAttachments) {
        final PublicServiceType publicServiceType = buildPublicServiceType(serviceType);
        final List<Attachment> attachmentMetadata = toAttachmentMetadata(encryptedAttachments);
        return buildMetadata(attachmentMetadata, data, publicServiceType);
    }

    public static Metadata buildMetadata(final List<Attachment> attachments, final Data data, final PublicServiceType publicServiceType) {
        final var contentStructure = new ContentStructure();
        contentStructure.setAttachments(attachments);
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);
        metadata.setPublicServiceType(publicServiceType);
        return metadata;
    }

    public static CreateSubmission buildSubmissionToAnnounce(final EncryptedSubmissionPayload submissionPayload) {
        final ServiceType serviceType = submissionPayload.getServiceType();
        final List<UUID> attachmentIds = new ArrayList<>(submissionPayload.getEncryptedAttachments().keySet());
        return createSubmission(submissionPayload.getDestinationId(), serviceType, attachmentIds);
    }

    public static SubmitSubmission buildSubmitSubmission(final UUID submissionId, final EncryptedSubmissionPayload submissionPayload) {
        final SubmitSubmission submission = new SubmitSubmission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData(submissionPayload.getEncryptedData());
        submission.setEncryptedMetadata(submissionPayload.getEncryptedMetadata());
        return submission;
    }

    public static SubmitSubmission buildSubmitSubmission(final UUID submissionId, final String encryptedData, final String encryptedMetadata) {
        final SubmitSubmission submission = new SubmitSubmission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);
        return submission;
    }

    private static Attachment toHashedAttachment(final AttachmentPayload attachmentPayload) {
        final var attachment = new Attachment();
        attachment.setAttachmentId(attachmentPayload.getAttachmentId());
        attachment.setPurpose(Purpose.ATTACHMENT);
        attachment.setFilename(attachmentPayload.getFile().getName());
        attachment.setMimeType(attachmentPayload.getMimeType());

        final var hash = new Hash();
        hash.setContent(attachmentPayload.getHashedData());
        hash.setSignatureType(SignatureType.SHA_512);
        attachment.setHash(hash);
        return attachment;
    }

    private static List<UUID> toAttachmentIds(final List<AttachmentPayload> attachmentPayloads) {
        return attachmentPayloads.stream()
                .map(AttachmentPayload::getAttachmentId)
                .collect(Collectors.toList());
    }

    public static List<Attachment> toAttachmentMetadata(final List<AttachmentPayload> attachmentPayloads) {
        return attachmentPayloads.stream()
                .map(SubmissionUtil::toHashedAttachment)
                .collect(Collectors.toList());
    }

    private static Function<Map.Entry<UUID, String>, AttachmentPayload> getEncryptedAttachmentPayload() {
        return attachment -> AttachmentPayload.builder()
                .encryptedData(attachment.getValue())
                .attachmentId(attachment.getKey())
                .build();
    }

    public static Function<File, AttachmentPayload> getFileAttachmentPayload() {
        return file -> AttachmentPayload.builder()
                .file(file)
                .attachmentId(UUID.randomUUID())
                .build();
    }

    public static List<AttachmentPayload> toAttachmentPayloads(final Map<UUID, String> encryptedAttachments) {
        return encryptedAttachments.entrySet()
                .stream()
                .map(getEncryptedAttachmentPayload())
                .collect(Collectors.toList());
    }

    public static CreateSubmission buildSubmissionToAnnounce(final UUID destinationId, final ServiceType serviceType, final List<AttachmentPayload> encryptedAttachments) {
        final List<UUID> attachmentIdsToAnnounce = toAttachmentIds(encryptedAttachments);
        return createSubmission(destinationId, serviceType, attachmentIdsToAnnounce);
    }

    public static List<SubmissionSchema> getSubmissionSchemasFromDestination(final Destination destination) {
        return destination.getServices().stream()
                .map(DestinationService::getSubmissionSchemas)
                .flatMap(Collection::stream)
                .collect(Collectors.toList());
    }

    public static Optional<URI> getSchemaUriForMimeType(final List<SubmissionSchema> submissionSchemas, final MimeType mimeType) {
        return submissionSchemas.stream()
                .filter(schema -> schema.getMimeType().equals(mimeType))
                .map(SubmissionSchema::getSchemaUri)
                .findFirst();
    }

    public static DecryptedAttachmentPayload buildDecryptedAttachmentPayload(final Attachment attachmentMetadata, final byte[] decryptedAttachment) {
        return DecryptedAttachmentPayload.builder()
                .decryptedContent(decryptedAttachment)
                .attachmentMetadata(attachmentMetadata)
                .build();
    }

    public static SentSubmission buildSentSubmission(final EncryptedSubmissionPayload submissionPayload, final List<AttachmentPayload> encryptedAttachments, final Submission submission) {
        return buildSentSubmission(encryptedAttachments, submissionPayload.getEncryptedData(), submissionPayload.getEncryptedMetadata(), submission);
    }

    public static SentSubmission buildSentSubmission(final List<AttachmentPayload> encryptedAttachments, final String encryptedData, final String encryptedMetadata, final Submission submission) {

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setData(getAuthTag(encryptedData));
        authenticationTags.setMetadata(getAuthTag(encryptedMetadata));
        authenticationTags.setAttachments(getAuthTagsFromAttachments(encryptedAttachments));

        return SentSubmission.builder()
                .submissionId(submission.getSubmissionId())
                .caseId(submission.getCaseId())
                .destinationId(submission.getDestinationId())
                .authenticationTags(authenticationTags)
                .build();
    }

    public static List<ReceivedAttachment> mapToReceivedAttachments(final List<DecryptedAttachmentPayload> decryptedAttachmentPayloads) {
        return decryptedAttachmentPayloads.stream()
                .map(SubmissionUtil::mapToReceivedAttachment)
                .collect(Collectors.toList());
    }

    public static Data createData(final SubmissionPayload submissionPayload, final String hashedData, final URI schemaUri) {
        final var hash = new Hash();
        hash.setContent(hashedData);
        hash.setSignatureType(SignatureType.SHA_512);

        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setMimeType(submissionPayload.getDataMimeType());
        submissionSchema.setSchemaUri(schemaUri);

        final var data = new Data();
        data.setSubmissionSchema(submissionSchema);
        data.setHash(hash);
        return data;
    }

    public static String getDataHashFromMetadata(final Metadata metadata) {
        return metadata.getContentStructure()
                .getData()
                .getHash()
                .getContent();
    }

    private static ReceivedAttachment mapToReceivedAttachment(final DecryptedAttachmentPayload payload) {
        final Attachment metadata = payload.getAttachmentMetadata();
        return ReceivedAttachment.builder()
                .attachmentId(metadata.getAttachmentId())
                .filename(metadata.getFilename())
                .mimeType(metadata.getMimeType())
                .description(metadata.getDescription())
                .data(payload.getDecryptedContent())
                .build();
    }

    private static Map<UUID, String> getAuthTagsFromAttachments(final List<AttachmentPayload> encryptedAttachments) {
        return encryptedAttachments.stream().collect(Collectors.toMap(AttachmentPayload::getAttachmentId, SubmissionUtil::getAuthTag));
    }

    private static String getAuthTag(final AttachmentPayload attachmentPayload) {
        return getAuthTag(attachmentPayload.getEncryptedData());
    }

    private static String getAuthTag(final String encryptedJWTData) {
        try {
            return JWEObject.parse(encryptedJWTData).getAuthTag().toString();
        } catch (final ParseException | NullPointerException e) {
            return "";
        }
    }

}
